/* cdpd_power.ps.c */                                                       
/* Custom received power model for radio link Transceiver Pipeline */

/****************************************/
/*            Michael Jiang             */
/*            08/10/1999                */
/****************************************/

#include <opnet.h>
#include <math.h>


/***** constants *****/

#define C					3.0E+08			/* speed of light (m/s) */
#define SIXTEEN_PI_SQ		157.91367		/* 16 times pi-squared */
#define FEET_PER_METER		3.3

#define TWO_PI				6.28
#define FADE_FAC			0.8		
#define NULLS_PER_SEC		50				/* 50 mph */
#define METERS_PER_MILE		1600

#define	RURAL				0
#define	SUBURBAN			1
#define	URBAN				2

#define	NO_FOLIAGE			0
#define	LOW_FOLIAGE			1
#define	MED_FOLIAGE			2
#define	HEAVY_FOLIAGE		3

/***** pipeline procedure *****/

void
cdpd_power (pkptr)
	Packet*		pkptr;
	{
	double		prop_distance, rcvd_power, path_loss;
	double		tx_power, tx_base_freq, tx_bandwidth, tx_center_freq;
	double		lambda, rx_ant_gain, tx_ant_gain;
	int			sig_lock;
	Objid		rx_ch_obid;
	double		in_band_tx_power, band_max, band_min;
	double		rx_base_freq, rx_bandwidth;
	double		h1, h2;
	double		cur_time;
	double		path_loss_intcpt, db_per_decade;
	int			setting_type, foliage_density;

	/** Compute the average power in Watts of the		**/
	/** signal associated with a transmitted packet.	**/
	FIN (cdpd_power (pkptr))

	/* If the incoming packet is 'valid', it may cause the receiver to	*/
	/* lock onto it. However, if the receiving node is disabled, then	*/
	/* the channel match should be set to noise.						*/
	if (op_td_get_int (pkptr, OPC_TDA_RA_MATCH_STATUS) == OPC_TDA_RA_MATCH_VALID)
		{
		if (op_td_is_set (pkptr, OPC_TDA_RA_ND_FAIL))
			{
			/* The receiving node is disabled.  Change	*/
			/* the channel match status to noise.		*/
			op_td_set_int (pkptr, OPC_TDA_RA_MATCH_STATUS, OPC_TDA_RA_MATCH_NOISE);
			}
		else
			{
			/* The receiving node is enabled.  Get	*/
			/* the address of the receiver channel.	*/
			rx_ch_obid = op_td_get_int (pkptr, OPC_TDA_RA_RX_CH_OBJID);
	
			/* If the receiver channel is already locked,		*/
			/* the packet will now be considered to be noise.	*/
			/* This prevents simultaneous reception of multiple	*/
			/* valid packets on any given radio channel.		*/
			op_ima_obj_attr_get (rx_ch_obid, "signal lock", &sig_lock);
			if (sig_lock)
				op_td_set_int (pkptr, OPC_TDA_RA_MATCH_STATUS, OPC_TDA_RA_MATCH_NOISE);
			else
				{
				/* Otherwise, the receiver channel will become	*/
				/* locked until the packet reception ends.		*/
				sig_lock = OPC_TRUE;
				op_ima_obj_attr_set (rx_ch_obid, "signal lock", sig_lock);
				}
			}
		}

	/* Get power allotted to transmitter channel. */
	tx_power = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_POWER);

	/* Get transmission frequency in Hz. */
	tx_base_freq = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_FREQ);
	tx_bandwidth = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_BW);
	tx_center_freq = tx_base_freq + (tx_bandwidth / 2.0);

	/* Caclculate wavelength (in meters). */
	lambda = C / tx_center_freq;

	/* Get distance between transmitter and receiver (in meters). */
	prop_distance = op_td_get_dbl (pkptr, OPC_TDA_RA_START_DIST);

	/* Compute the path loss for this distance and wavelength. */
	if (prop_distance > 0.0)
		{
		path_loss = (lambda * lambda) / 
			(SIXTEEN_PI_SQ * prop_distance * prop_distance);
		}
	else
		path_loss = 1.0;

	/* Determine the receiver bandwidth and base frequency. */
	rx_base_freq = op_td_get_dbl (pkptr, OPC_TDA_RA_RX_FREQ);
	rx_bandwidth = op_td_get_dbl (pkptr, OPC_TDA_RA_RX_BW);

	/* Use these values to determine the band overlap with the transmitter.	*/
	/* Note that if there were no overlap at all, the packet would already	*/
	/* have been filtered by the channel match stage.						*/

	/* The base of the overlap band is the highest base frequency. */
	if (rx_base_freq > tx_base_freq) 
		band_min = rx_base_freq;
	else
		band_min = tx_base_freq;
	
	/* The top of the overlap band is the lowest end frequency. */
	if (rx_base_freq + rx_bandwidth > tx_base_freq + tx_bandwidth)
		band_max = tx_base_freq + tx_bandwidth;
	else
		band_max = rx_base_freq + rx_bandwidth;

	/* Compute the amount of in-band transmitter power. */
	in_band_tx_power = tx_power * (band_max - band_min) / tx_bandwidth;

	/* Get antenna gains in dB. */
	tx_ant_gain = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_GAIN);
	rx_ant_gain = op_td_get_dbl (pkptr, OPC_TDA_RA_RX_GAIN);

	/* Get current sim time. */
	cur_time = op_sim_time ();

	/* Add fade zone characteristics to in band power. */
	in_band_tx_power = in_band_tx_power - ((FADE_FAC * in_band_tx_power) * (sin (TWO_PI * cur_time * NULLS_PER_SEC)));

	/* Convert in-band transmitted power to decibels. */
	in_band_tx_power = 10.0 * log10 (in_band_tx_power * 1000.0);

	/* Get the heights of the 2 antennae. */
	h1 = op_td_get_dbl (pkptr, OPC_TDA_RA_TX_ALT);
	h2 = op_td_get_dbl (pkptr, OPC_TDA_RA_RX_ALT);

	/* Convert heights from meters to feet. */
	h1 = h1 * FEET_PER_METER;
	h2 = h2 * FEET_PER_METER;

	/* Convert prop_distance to miles. */
	prop_distance = prop_distance / METERS_PER_MILE;

	/* Find out what type of area we are in and setup the path loss */
	/* constants accordingly, which is specified by the simulation  */	/* attribute setting_type. Get the value of this attribute.     */
	/* First set the variable to the default value and then         */
	/* overwrite it, if another value is explicitly set for the     */
	/* simulation attribute.                                        */
	setting_type = URBAN;
	if (op_ima_sim_attr_exists ("setting_type") == OPC_TRUE)
	    {
	    op_ima_sim_attr_get (OPC_IMA_INTEGER, "setting_type", &setting_type);
	    }

	/* Get the value of simulation attribute foliage coverage using */
	/* same approach.                                               */
	foliage_density = NO_FOLIAGE;
	if (op_ima_sim_attr_exists ("foliage_density") == OPC_TRUE)
	    {
	    op_ima_sim_attr_get (OPC_IMA_INTEGER, "foliage_density", &foliage_density);
	    }

	/* Normal settings for > 1.0 mile distance. */
	if(setting_type == RURAL)
		{
		path_loss_intcpt = -49.0;
		db_per_decade = 43.5;
		}

	else if (setting_type == SUBURBAN)
		{
		path_loss_intcpt = -61.7;
		db_per_decade = 40.0;
		}

	else if (setting_type == URBAN)
		{
		path_loss_intcpt = -75.0;
		db_per_decade = 37.0;
		}
		
	/* Characteristics change in close to the tower. */
	if (prop_distance < 1.0)
		{
		db_per_decade = 24.0;
		}

	if (foliage_density == LOW_FOLIAGE)
		db_per_decade = db_per_decade + 1.0;

	else if (foliage_density == MED_FOLIAGE)
		db_per_decade = db_per_decade + 3.0;

	else if (foliage_density == HEAVY_FOLIAGE)
		db_per_decade = db_per_decade + 5.0;

	/* Calculate received power according to Eq. 4.2-19 (in dBm) */
    /* (based on the analysis done by "William C. Y. Lee" in the */
	/* book "Mobile Cellular Telecommunications Systems").       */
	rcvd_power = in_band_tx_power - 96 + path_loss_intcpt - (db_per_decade * log10(prop_distance));
	rcvd_power = rcvd_power + (20 * log10(h1)) + (10 * log10(h2));
	rcvd_power = rcvd_power + tx_ant_gain + rx_ant_gain;

	/* Assign the received power level (in dB) */
	op_td_set_dbl (pkptr, OPC_TDA_RA_RCVD_POWER, rcvd_power);

	FOUT
	}
